@@ -5,6 +5,8 @@ import numpy as np
 import matplotlib.animation as animation
 import pandas as pd
 
+_FORBIDDEN_SUFFIXES = ['_acc', '__hist', '__futr']
+
 
 class _ConstantsDict(_collections.MutableMapping, dict):
     """A custom dictionary for storing constants."""
@@ -24,6 +26,10 @@ class _ConstantsDict(_collections.MutableMapping, dict):
             msg = ('{} is a reserved parameter name. '
                    'Choose something different.')
             raise ValueError(msg.format(key))
+        elif any([key.endswith(s) for s in _FORBIDDEN_SUFFIXES]):
+            msg = ('{} are reserved suffixes. '
+                   'Choose something different.')
+            raise ValueError(msg.format(_FORBIDDEN_SUFFIXES))
         else:
             dict.__setitem__(self, key, value)
 
@@ -57,12 +63,8 @@ class _CoordinatesDict(_collections.MutableMapping, dict):
         return dict.__getitem__(self, key)
 
     def __setitem__(self, key, value):
-        if len(self.keys()) == 1 and key != list(self.keys())[0]:
-            msg = ("There is already a coordinate set for this system, only "
-                   "one coordinate is permitted. Use del to remove the "
-                   "coordinate if you'd like to add a new one.")
-            raise ValueError(msg)
-        elif not key.isidentifier():
+        # TODO : Shouldn't allow coordinates to be named with suffix _vel.
+        if not key.isidentifier():
             msg = ('{} is not a valid coordinate or speed name. '
                    'It must be a valid Python variable name.')
             raise ValueError(msg.format(key))
@@ -70,6 +72,10 @@ class _CoordinatesDict(_collections.MutableMapping, dict):
             msg = ('{} is a reserved parameter name. '
                    'Choose something different.')
             raise ValueError(msg.format(key))
+        elif any([key.endswith(s) for s in _FORBIDDEN_SUFFIXES]):
+            msg = ('{} are reserved suffixes. '
+                   'Choose something different.')
+            raise ValueError(msg.format(_FORBIDDEN_SUFFIXES))
         else:
             dict.__setitem__(self, key, value)
 
@@ -86,32 +92,64 @@ class _CoordinatesDict(_collections.MutableMapping, dict):
         return dict.__contains__(self, x)
 
 
+class _SingleDoFCoordinatesDict(_CoordinatesDict):
+    def __setitem__(self, key, value):
+        if len(self.keys()) == 1 and key != list(self.keys())[0]:
+            msg = ("There is already a coordinate set for this system, only "
+                   "one coordinate is permitted. Use del to remove the "
+                   "coordinate if you'd like to add a new one.")
+            raise ValueError(msg)
+        else:
+            super(_SingleDoFCoordinatesDict, self).__setitem__(key, value)
+
+
 class _MeasurementsDict(_collections.MutableMapping, dict):
 
+    def _check_for_duplicate_keys(self):
+        c_keys = list(self._constants.keys())
+        g_keys = list(self._coordinates.keys())
+        s_keys = list(self._speeds.keys())
+        m_keys = list(self.keys())
+        all_keys = c_keys + g_keys + s_keys + m_keys
+        if len(set(all_keys)) < len(all_keys):
+            dups = []
+            sorted_keys = sorted(all_keys)
+            for i, k in enumerate(sorted_keys[1:]):
+                if k == sorted_keys[i - 1]:
+                    dups.append(k)
+            msg = ("{} are duplicate keys in your system's parameters. "
+                   "Duplicates are not allowed.")
+            raise KeyError(msg.format(dups))
+
     def _compute_value(self, key):
 
         func = self._funcs[key]
 
         def get_par(k):
-            try:
+            if k == 'time':
+                v = self._time['t']
+            elif k in self._constants:
                 v = self._constants[k]
-            except KeyError:
-                try:
-                    v = self._coordinates[k]
-                except KeyError:
-                    try:
-                        v = self._speeds[k]
-                    except KeyError:
-                        v = self[k]
+            elif k in self._coordinates:
+                v = self._coordinates[k]
+            elif k in self._speeds:
+                v = self._speeds[k]
+            elif k in self:
+                v = self[k]
+            else:
+                msg = ("{} not in constants, coordinates, speeds, or "
+                       "measurements.")
+                raise KeyError(msg.format(k))
             return v
 
-        # TODO : getargspec is deprecated, supposedly signature cna do the same
+        # TODO : getargspec is deprecated, supposedly signature can do the same
         # thing but the args are in a dictionary and it isn't clear to me they
         # are ordered.
         args = [get_par(k) for k in getargspec(func).args]
         return func(*args)
 
     def __getitem__(self, key):
+        self._check_for_duplicate_keys()
         return self._compute_value(key)
 
     def __setitem__(self, key, value):
@@ -120,7 +158,7 @@ class _MeasurementsDict(_collections.MutableMapping, dict):
         raise ValueError(msg)
 
     def __delitem__(self, key):
-        # TODO : Should also remove from self._funcs
+        dict.__delitem__(self._funcs, key)
         dict.__delitem__(self, key)
 
     def __iter__(self):
@@ -134,47 +172,9 @@ class _MeasurementsDict(_collections.MutableMapping, dict):
 
 
 class System(object):
-    """This is the base system for any single or multi degree of freedom
-    systems. It can be subclassed to make a custom system.
-
-    Attributes
-    ==========
-    constants : _ConstantsDict
-        A custom dictionary that contains parameters that do not vary with
-        time.
-    coordinates : _CoordinatesDict
-        A custom dictionary that contains the generalized coordinate which
-        varies with time.
-    speeds : _CoordinatesDict
-        A custom dictionary that contains the generalized speed which varies
-        with time.
-    measurements : _MeasurementsDict
-        A custom dictionary that contains parameters that are functions of the
-        constants, coordinates, and other measurements.
-    config_plot_func : function
-        A function that generates a matplotlib plot that uses the instantaneous
-        values of the constants, coordinates, and measurements.
-    config_plot_update_func : function
-        A function that updates the configuration plot that uses the time
-        series values of the constants, coordinates, and measurements. Defines
-        a matplotlib animation frame.
-
-    Methods
-    =======
-    add_measurement
-        Used to dynamically add functions to compute measurement values.
-    free_response
-        Simulates the system and returns a time series of the coordinates and
-        measurments.
-    plot_configuration
-        Generates the plot defined by ``config_plot_func``.
-    animate_configutation
-        Generates the animation defined by ``config_plot_func`` and
-        ``config_plot_update_func``.
-    period
-        Returns the damped natural period of the system.
-
-    """
+    """This is the abstract base class for any single or multi degree of
+    freedom system. It can be sub-classed to make a custom system or the
+    necessary methods can be added dynamically."""
 
     _time_var_name = 'time'
     _vel_append = '_vel'
@@ -182,7 +182,9 @@ class System(object):
 
     def __init__(self):
 
-        # TODO : Allow constants, coords, and meas to be set on intialization.
+        # TODO : Allow constants, coords, and meas to be set on initialization.
+
+        self._time = {'t': 0.0}  # this needs to be a mutable object
 
         self._constants = _ConstantsDict({})
         self._coordinates = _CoordinatesDict({})
@@ -192,6 +194,7 @@ class System(object):
         self._measurements._constants = self._constants
         self._measurements._coordinates = self._coordinates
         self._measurements._speeds = self._speeds
+        self._measurements._time = self._time
         self._measurements._funcs = {}
 
         self._config_plot_func = None
@@ -199,7 +202,8 @@ class System(object):
 
     @property
     def constants(self):
-        """A dictionary containing the all of the system's constants.
+        """A dictionary containing the all of the system's constants, i.e.
+        parameters that do not vary with time.
 
         Examples
         ========
@@ -228,8 +232,9 @@ class System(object):
 
     @property
     def coordinates(self):
-        """A dictionary containing the system's generalized coordinates. These
-        values will be used as initial conditions in simulations.
+        """A dictionary containing the system's generalized coordinates, i.e.
+        coordinate parameters that vary with time. These values will be used as
+        initial conditions in simulations.
 
         Examples
         ========
@@ -249,7 +254,9 @@ class System(object):
 
     @property
     def speeds(self):
-        """A dictionary containing the system's generalized speed.
+        """A dictionary containing the system's generalized speeds, i.e. speed
+        parameters that vary with time. These values will be used as initial
+        conditions in simulations.
 
         Examples
         ========
@@ -296,7 +303,9 @@ class System(object):
 
     @property
     def measurements(self):
-        """A dictionary containing the all of the system's measurements."""
+        """A dictionary containing the all of the system's measurements, i.e.
+        parameters that are functions of the constants, coordinates, speeds,
+        and other measurements."""
         return self._measurements
 
     @measurements.setter
@@ -383,25 +392,22 @@ class System(object):
     def _get_par_vals(self, par_name):
         """Returns the value of any variable stored in the parameters,
         coordinates, or measurements dictionaries."""
-        # TODO : Maybe just merging the three dicts is better? What to do about
-        # name clashes in the dictionaries? Garbage in garbage out?
-        # TODO : Raise useful error message if par_name not in any of the
-        # dicts.
+        self._measurements._check_for_duplicate_keys()
+        # TODO : This is duplicate of similar code in
+        # _MeasurementsDict._compute_value(). Shouldn't have this redundancy.
         msg = '{} is not in constants, coordinates, speeds, or measurements.'
-        try:
-            v = self.constants[par_name]
-        except KeyError:
-            try:
-                v = self.coordinates[par_name]
-            except KeyError:
-                try:
-                    v = self.speeds[par_name]
-                except KeyError:
-                    try:
-                        v = self.measurements[par_name]
-                    except KeyError:
-                        raise KeyError(msg.format(par_name))
-        return v
+        if par_name.lower() == 'time':
+            return self._time['t']
+        elif par_name in self.constants:
+            return self.constants[par_name]
+        elif par_name in self.coordinates:
+            return self.coordinates[par_name]
+        elif par_name in self.speeds:
+            return self.speeds[par_name]
+        elif par_name in self._measurements:
+            return self.measurements[par_name]
+        else:
+            raise KeyError(msg.format(par_name))
 
     def add_measurement(self, name, func):
         """Creates a new measurement entry in the measurements attribute that
@@ -416,7 +422,7 @@ class System(object):
             This function must only have existing parameter, coordinate, or
             measurement names in the function signature. These can be a subset
             of the available choices and any order is permitted. The function
-            must be able to operate on arrys, i.e. use NumPy vectorized
+            must be able to operate on arrays, i.e. use NumPy vectorized
             functions inside. It should return a single variable, scalar or
             array, that gives the values of the measurement.
 
@@ -440,6 +446,10 @@ class System(object):
             msg = ('{} is a reserved parameter name. '
                    'Choose something different.')
             raise ValueError(msg.format(name))
+        elif any([name.endswith(s) for s in _FORBIDDEN_SUFFIXES]):
+            msg = ('{} are reserved suffixes. '
+                   'Choose something different.')
+            raise ValueError(msg.format(_FORBIDDEN_SUFFIXES))
         elif name in (list(self.constants.keys()) +
                       list(self.coordinates.keys()) +
                       list(self.speeds.keys())):
@@ -453,42 +463,64 @@ class System(object):
 
     def _state_traj_to_dataframe(self, times, pos, vel, acc):
 
-        coord_name = list(self.coordinates.keys())[0]
-        speed_name = coord_name + self._vel_append
+        pos = np.atleast_2d(pos)
+        vel = np.atleast_2d(vel)
+        acc = np.atleast_2d(acc)
 
-        # TODO : What if they added a coordinate with the vel or acc names?
-        df = pd.DataFrame({coord_name: pos,
-                           speed_name: vel,
-                           coord_name + self._acc_append: acc},
-                          index=times)
+        # TODO : What if they added a coordinate with the acc names?
+        data = {}
+        for i, c_name in enumerate(list(self.coordinates.keys())):
+            data[c_name] = pos[i]
+            data[c_name + self._acc_append] = acc[i]
+        for i, s_name in enumerate(list(self.speeds.keys())):
+            data[s_name] = vel[i]
+        df = pd.DataFrame(data, index=times)
         df.index.name = self._time_var_name
 
         # TODO : Allow acc to be used in measurements.
-        # TODO : Need a way to compute the measurements as array values based
-        # on the coordinate changing at each time but the current method of
-        # letting the measurement be computed by the stored coordinate scalar
-        # is a bit problematic.
-        for k, v in self.measurements.items():
-            vals = np.zeros_like(times)
-            x0 = list(self.coordinates.values())[0]
-            v0 = list(self.speeds.values())[0]
-            for i, (xi, vi) in enumerate(zip(pos, vel)):
-                self.coordinates[coord_name] = xi
-                self.speeds[speed_name] = vi
-                vals[i] = self.measurements[k]
-            self.coordinates[coord_name] = x0
-            self.speeds[speed_name] = v0
-            df[k] = vals
+        # store current values of coords, speeds, and time
+        x0s = list(self.coordinates.values())
+        v0s = list(self.speeds.values())
+        t0 = self._time['t']
+
+        # set coord, speeds, and time to arrays
+        for i, c_name in enumerate(list(self.coordinates.keys())):
+            self.coordinates[c_name] = pos[i]
+        for i, s_name in enumerate(list(self.speeds.keys())):
+            self.speeds[s_name] = vel[i]
+        self._time['t'] = times
+
+        # compute each measurement
+        for name, value in self.measurements.items():
+            df[name] = value
+
+        # set the coords, speeds, and time back to original values
+        for i, c_name in enumerate(list(self.coordinates.keys())):
+            self.coordinates[c_name] = x0s[i]
+        for i, s_name in enumerate(list(self.speeds.keys())):
+            self.speeds[s_name] = v0s[i]
+        self._time['t'] = t0
 
         return df
 
+    @staticmethod
+    def _calc_times(final_time, initial_time, sample_rate):
+        # TODO : Should have the option to pass in unequally spaced monotonic
+        # time arrays.
+
+        if final_time < initial_time:
+            raise ValueError('Final time must be greater than initial time.')
+
+        delta = final_time - initial_time
+        num = int(round(sample_rate * delta)) + 1
+        return np.linspace(initial_time, final_time, num=num)
+
     def free_response(self, final_time, initial_time=0.0, sample_rate=100,
                       **kwargs):
-        """Returns a Pandas data frame with monotonic time values as the index
-        and columns for each coordinate and measurement at the time value for
-        that row. Note that this data frame is stored on the system as the
-        vairable ``result`` until this method is called again, which will
-        overwrite it.
+        """Returns a data frame with monotonic time values as the index and
+        columns for each coordinate and measurement at the time value for that
+        row. Note that this data frame is stored on the system as the variable
+        ``result`` until this method is called again, which will overwrite it.
 
         Parameters
         ==========
@@ -511,15 +543,8 @@ class System(object):
             measurements as columns.
 
         """
-        # TODO : Should have the option to pass in unequally spaced monotonic
-        # time arrays.
 
-        if final_time < initial_time:
-            raise ValueError('Final time must be greater than initial time.')
-
-        delta = final_time - initial_time
-        num = int(round(sample_rate * delta)) + 1
-        times = np.linspace(initial_time, final_time, num=num)
+        times = self._calc_times(final_time, initial_time, sample_rate)
 
         pos, vel, acc = self._generate_state_trajectories(times, **kwargs)
 
@@ -543,30 +568,90 @@ class System(object):
         # TODO : Most plots in pandas, etc return the axes not the figure. I
         # think the parent figure can always be gotten from an axis.
         if self.config_plot_func is None:
-            msg = 'No ploting function has been assigned to config_plot_func.'
-            raise AttributeError(msg)
+            msg = 'No plotting function has been assigned to config_plot_func.'
+            raise ValueError(msg)
         else:
             args = []
             for k in getargspec(self.config_plot_func).args:
                 if k == 'time':
-                    args.append(0.0)  # static config defaults to t=0.0
+                    args.append(self._time['t'])
                 elif k == 'time__hist':
-                    args.append(0.0)  # static config defaults to t=0.0
+                    args.append(self._time['t'])
+                elif k == 'time__futr':
+                    args.append(self._time['t'])
                 elif k.endswith('__hist'):
                     args.append(self._get_par_vals(k[:-6]))
+                elif k.endswith('__futr'):
+                    args.append(self._get_par_vals(k[:-6]))
                 else:
                     args.append(self._get_par_vals(k))
             return self.config_plot_func(*args)
 
-    def animate_configuration(self, **kwargs):
+    def _resample_trajectories(self, sample_rate=60):
+
+        trajectories = self.result.copy()
+
+        new_times = self._calc_times(self.result.index[-1],
+                                     self.result.index[0], sample_rate)
+        # first and last will be same, so drop
+        new_times_trunc = new_times[1:-1]
+
+        num_cols = len(trajectories.columns)
+
+        missing_vals = np.nan * np.ones((len(new_times_trunc), num_cols))
+
+        nan_df = pd.DataFrame(missing_vals, columns=trajectories.columns,
+                              index=new_times_trunc)
+
+        df_with_missing = pd.concat((trajectories, nan_df)).sort_index()
+
+        interpolated_df = df_with_missing.interpolate(method='index')
+
+        trajectories = interpolated_df.loc[new_times]
+
+        interval = 1000 / 60  # milliseconds
+
+        return trajectories, interval
+
+    def animate_configuration(self, fps=30, **kwargs):
         """Returns a matplotlib animation function based on the configuration
-        plot and the configuration plot update function."""
+        plot and the configuration plot update function.
+
+        Parameters
+        ==========
+        fps : integer
+            The frames per second that should be displayed in the animation.
+            The latest trajectory will be resampled via linear interpolation to
+            create the correct number of frames. Note that the frame rate will
+            depend on the CPU speed of the computer. You'll likely have to
+            adjust this by trial and error to get something that matches well
+            for your computer if you want the animation to run in real time.
+        **kwargs
+            Any extra keyword arguments will be passed to
+            ``matplotlib.animation.FuncAnimation()``. The ``interval`` keyword
+            argument will be ignored.
+
+        """
 
         if self.config_plot_update_func is None:
             msg = ('No ploting update function has been assigned to '
                    'config_plot_update_func.')
+            raise ValueError(msg)
+
+        kwargs.pop('interval', None)  # ignore the user's supplied interval
+        try:
+            sample_rate = int(1.0 / np.diff(self.result.index).mean())
+        except AttributeError:
+            msg = ("No trajectory has been computed yet, so the animation "
+                   "can't run. Run one of the response functions.")
             raise AttributeError(msg)
 
+        fps = int(fps)
+        if sample_rate != fps:
+            trajectories, interval = self._resample_trajectories(fps)
+        else:
+            trajectories, interval = self.result, 1000 / sample_rate
+
         # TODO : Could be:
         # axes, *objs_to_modify = ..
         # try:
@@ -590,16 +675,36 @@ class System(object):
                 if k == 'time':
                     args.append(time)
                 elif k == 'time__hist':
-                    args.append(self.result[:time].index)
+                    args.append(trajectories[:time].index)
+                elif k == 'time__futr':
+                    args.append(trajectories[time:].index)
                 elif k.endswith('__hist'):
-                    args.append(self.result[k[:-6]][:time])
-                else:
-                    try:
-                        args.append(row[k])
-                    except KeyError:
-                        # requires these to be in the same order
-                        args.append(pop_list.pop(0))
+                    args.append(trajectories[k[:-6]][:time])
+                elif k.endswith('__futr'):
+                    args.append(trajectories[k[:-6]][time:])
+                elif k in trajectories:  # constants, coordinates, measurements
+                    args.append(row[k])
+                elif k in self.constants:
+                    args.append(self.constants[k])
+                else:  # must be matplotlib object
+                    # TODO : This last part is quite fragile. It requires these
+                    # remaining args to be in the same order as the returned
+                    # tuple from the plot function and there is no way to know
+                    # if these are the correct objects to append if the order
+                    # isn't correct.
+                    args.append(pop_list.pop(0))
             self.config_plot_update_func(*args)
 
-        return animation.FuncAnimation(fig, gen_frame, fargs=(objs_to_modify, ),
-                                       frames=self.result.iterrows(), **kwargs)
+        # NOTE : This is useful to uncomment in debugging because the errors
+        # push to the top if in the FuncAnimation.
+        #gen_frame((1.0, self.result.iloc[0]), list(objs_to_modify))
+
+        # NOTE : If the iterrows() generator is not converted to a list then
+        # matplotlib will throw a StopIteration error when the animation
+        # reaches the last frame instead of repeating. This causes headaches in
+        # the notebook and elsewhere. See issue #39 in the resonance repo.
+        return animation.FuncAnimation(fig, gen_frame,
+                                       fargs=(objs_to_modify, ),
+                                       frames=list(trajectories.iterrows()),
+                                       interval=interval,
+                                       **kwargs)
